@@ -1,15 +1,12 @@
-use strict 'vars', 'subs';  # Moved to stop Perl::Critic carping when Dist::Zilla adds
-                            # $VERSION information.
+use strict 'vars', 'subs';                                                                                                         # Moved to stop Perl::Critic carping when Dist::Zilla adds
+                                                                                                                                   # $VERSION information.
 
 package DBIx::Squirrel;
-$DBIx::Squirrel::VERSION = '1.1.1'; # TRIAL
+$DBIx::Squirrel::VERSION = '1.2.1'; # TRIAL
 use warnings;
 use constant E_BAD_ENT_BIND     => 'Cannot associate with an invalid object';
 use constant E_EXP_HASH_ARR_REF => 'Expected a reference to a HASH or ARRAY';
 
-use Scalar::Util 'reftype';
-use Sub::Name;
-
 =pod
 
 =encoding UTF-8
@@ -20,7 +17,7 @@ DBIx::Squirrel - A module for working with databases
 
 =head1 VERSION
 
-version 1.1.1
+version 1.2.1
 
 =head1 SYNOPSIS
 
@@ -34,11 +31,10 @@ version 1.1.1
     $itr = $sth->iterate('1001099');
     while ($row = $itr->next) {...}
 
-    # Or, use it and have it create and import helper functions that you
-    # can define at runtime use (and reuse) to interact with database
-    # connections, statements and iterators.
+    # Or, use it and have it create and import helper functions that
+    # you can use to interact with database objects.
 
-    use DBIx::Squirrel 'db', 'st', 'it';
+    use DBIx::Squirrel database_objects=>['db', 'st', 'it'];
 
     db DBIx::Squirrel->connect($dsn, $user, $pass, \%attr);
     st db->prepare('SELECT * FROM product WHERE id = ?');
@@ -212,7 +208,11 @@ version 1.1.1
 
 =cut
 
-use DBI                ();
+use DBI;
+use Exporter;
+use Scalar::Util 'reftype';
+use Sub::Name;
+
 use DBIx::Squirrel::dr ();
 use DBIx::Squirrel::db ();
 use DBIx::Squirrel::st ();
@@ -222,11 +222,13 @@ use DBIx::Squirrel::rc ();
 use DBIx::Squirrel::util 'throw';
 
 BEGIN {
-    *err    = *DBI::err;
-    *errstr = *DBI::errstr;
-    *rows   = *DBI::rows;
-    *lasth  = *DBI::lasth;
-    *state  = *DBI::state;
+    *EXPORT_OK   = *DBI::EXPORT_OK;
+    *EXPORT_TAGS = *DBI::EXPORT_TAGS;
+    *err         = *DBI::err;
+    *errstr      = *DBI::errstr;
+    *rows        = *DBI::rows;
+    *lasth       = *DBI::lasth;
+    *state       = *DBI::state;
 
     *connect         = *DBIx::Squirrel::dr::connect;
     *connect_cached  = *DBIx::Squirrel::dr::connect_cached;
@@ -243,94 +245,95 @@ BEGIN {
     *NORMALIZE_SQL = *NORMALISE_SQL;
 }
 
-# By appending a list of one or more names to the caller's "use" directive,
-# can have the DBIx::Squirrel package define helper functions ("helpers")
-# during Perl's compile-phase. All listed helpers are exported to the
-# caller's namespace. Any helper that has already been defined will only
-# be exported; it will not be re-defined.
-#
-# A helper is a callable name that may be associated with a database
-# object (connection, statement, iterator) during runtime. A helper acts
-# as a shim, allowing the caller to get (or address) the underlying object
-# without Perl's traditional line-noise ("$", "->", "(...)").
-#
-# Absent any arguments, a call to a helper simply returns the reference
-# to the object itself. When called with arguments, a statement or
-# iterator helper will behave as if a call was made to the underlying
-# object's "execute" method. A database connection helper doesn't care
-# what is passed; a reference to the connection is all that is returned.
-#
-# Since some statements do not require arguments, execution is coerced
-# by passing a reference to an empty ARRAY or HASH. For the sake of
-# consistency, a reference to an ARRAY or HASH containing arguments
-# may also be passed.
-
 sub import {
     my $class  = shift;
     my $caller = caller;
 
-    my @helper_names = @_;
+    my ( @helpers, @dbi_imports );
+    while (@_) {
+        if ( $_[0] =~ m/^database_objects?$/i ) {
+            shift;
 
-    for my $name (@helper_names) {
-        my $symbol = $class . '::' . $name;
-
-        # Define the symbol once only!
-
-        unless ( defined &{$symbol} ) {
-            *{$symbol} = subname(
-                $symbol => sub {
-                    if (@_) {
-
-                        # No reason NOT to have helpers act as proxies for
-                        # DBI connection and statement objects, too!
-
-                        if (   UNIVERSAL::isa( $_[0], 'DBI::db' )
-                            or UNIVERSAL::isa( $_[0], 'DBI::st' )
-                            or UNIVERSAL::isa( $_[0], 'DBIx::Squirrel::it' ) )
-                        {
-                            ${$symbol} = shift;
-                            return ${$symbol};
+            if (@_) {
+                if ( defined $_[0] ) {
+                    if ( ref $_[0] ) {
+                        if ( reftype( $_[0] ) eq 'ARRAY' ) {
+                            push @helpers, @{ +shift };
                         }
-
-                        throw E_BAD_ENT_BIND;
+                    } else {
+                        push @helpers, shift;
                     }
-
-                    # Return nothing if no association is defined!
-
-                    return unless defined ${$symbol};
-
-                    # If we have arguments remaining and an object we can
-                    # meaningfully address, dispatch the "execute" method!
-
-                    if (@_
-                        and (  UNIVERSAL::isa( ${$symbol}, 'DBI::st' )
-                            or UNIVERSAL::isa( ${$symbol}, 'DBIx::Squirrel::it' ) )
-                      )
-                    {
-                        my @params = do {
-                            if ( @_ == 1 && ref $_[0] ) {
-                                if ( reftype( $_[0] ) eq 'ARRAY' ) {
-                                    @{ +shift };
-                                } elsif ( reftype( $_[0] ) eq 'HASH' ) {
-                                    %{ +shift };
-                                } else {
-                                    throw E_EXP_HASH_ARR_REF;
-                                }
-                            } else {
-                                @_;
-                            }
-                        };
-
-                        return ${$symbol}->execute(@params);
-                    }
-
-                    # For any other situation, return a reference to the
-                    # associated object!
-
-                    return ${$symbol};
                 }
-            );
+            }
+        } else {
+            push @dbi_imports, shift;
         }
+    }
+
+    my %seen;
+    @helpers = grep { !$seen{$_}++ } @helpers;
+
+    for my $name (@helpers) {
+        my $symbol = $class . '::' . $name;
+        my $helper = sub {
+
+            # First time the helper receives a value then it is assumed to be
+            # the reference to the associated database object.
+
+            unless ( defined ${$symbol} ) {
+                if (@_) {
+                    if (   UNIVERSAL::isa( $_[0], 'DBI::db' )
+                        or UNIVERSAL::isa( $_[0], 'DBI::st' )
+                        or UNIVERSAL::isa( $_[0], 'DBIx::Squirrel::it' ) )
+                    {
+                        ${$symbol} = shift;
+                        return ${$symbol};
+                    }
+
+                    throw E_BAD_ENT_BIND;
+                }
+            }
+
+            # At this point, return nothing if no association is defined.
+
+            return unless defined ${$symbol};
+
+
+            # At this point, we have an association. If we have arguments then
+            # we are addressing the associated database object, and they are
+            # passed to the object's relevant method for processing.
+
+            if (@_) {
+                my @params = do {
+                    if ( @_ == 1 && ref $_[0] ) {
+                        if ( reftype( $_[0] ) eq 'ARRAY' ) {
+                            @{ +shift };
+                        } elsif ( reftype( $_[0] ) eq 'HASH' ) {
+                            %{ +shift };
+                        } else {
+                            throw E_EXP_HASH_ARR_REF;
+                        }
+                    } else {
+                        @_;
+                    }
+                };
+
+                if ( UNIVERSAL::isa( ${$symbol}, 'DBI::db' ) ) {
+                    return ${$symbol}->prepare(@params);
+                } elsif ( UNIVERSAL::isa( ${$symbol}, 'DBI::st' ) ) {
+                    return ${$symbol}->execute(@params);
+                } elsif ( UNIVERSAL::isa( ${$symbol}, 'DBIx::Squirrel::it' ) ) {
+                    return ${$symbol}->iterate(@params);
+                }
+            }
+
+            # At this point, return the reference to the associated database
+            # object.
+
+            return ${$symbol};
+        };
+
+        *{$symbol} = subname( $name => $helper );
 
         # Export any relevant symbols to caller's namespace.
 
@@ -339,26 +342,22 @@ sub import {
         }
     }
 
-    return $class;
-}
+    # If the caller tried to import any of DBI's exports then take care
+    # of that, otherwise just return our class name.
 
-# Provide a means for undefining and unexporting helpers.
+    if (@dbi_imports) {
 
-sub unimport {
-    my $class  = shift;
-    my $caller = caller;
+        # First import them here.
 
-    my @helper_names = @_;
+        DBI->import(@dbi_imports);
 
-    for my $name (@helper_names) {
-        my $symbol = $class . '::' . $name;
+        # Then export them to the caller!
 
-        undef &{$symbol} if defined &{$symbol};
-        undef ${$symbol} if defined ${$symbol};
-        undef &{ $caller . '::' . $name } if defined &{ $caller . '::' . $name };
+        @_ = ( 'DBIx::Squirrel', @dbi_imports );
+        goto &Exporter::import;
+    } else {
+        return $class;
     }
-
-    return $class;
 }
 
 1;
@@ -370,61 +369,142 @@ __END__
 
 =head1 DESCRIPTION
 
-The C<DBIx::Squirrel> package extends the C<DBI>, offering a
-few useful conveniences over and above what its ancestor already
-provides.
+The C<DBIx::Squirrel> package extends the C<DBI>, by providing a few extra
+conveniences that are subtle and additive in nature, and, hopefully, quite
+useful.
 
-The enhancements provided by C<DBIx::Squirrel> are subtle, and
-they are additive in nature.
+=head2 Importing the package
 
-=head3 Importing the package
+In the simplest case, just import the package as you would any other:
 
-Simply use the package as you would any other:
+    use DBIx::Squirrel;
 
-    use DBIx::Squirrel [LIST-OF-IMPORTS];
+Any symbols and tags that you would typically import from the C<DBI> can
+also be requested via C<DBIx::Squirrel>:
 
-The optional list of imports may contain the names of special proxy
-functions that are imported into the caller's namespace.
+    use DBIx::Squirrel DBI-IMPORT-LIST;
 
-Proxy functions are simply syntactic sugar that you can associate with
-and use to address database, statement and/or iterator objects during
-runtime. 
+If required (and in addition to any C<DBI> imports), C<DBIx::Squirrel> can
+create and import Database Object Helper functions for you:
 
-What you call these functions is entirely up to you.
+    use DBIx::Squirrel database_object=>NAME;
+    use DBIx::Squirrel database_objects=>[NAMES];
 
-=head4 Proxy functions
+=head3 Database Object Helper Functions
 
-    use DBIx::Squirrel qw(db named_foo all_foo);
+A database object helper is nothing more than a standard function providing
+some syntactic sugar in the form of a polymorphic interface for interacting
+with database entities such as database connections, statements and
+iterators.
 
-    # Define the association between the "db" proxy function
-    # and our database connection...
+While it is not absolutely necessary to use themE<mdash>you could just as
+easily use scalar referencesE<mdash>helper functions do possess the advantage
+of being shared more easily among package namespaces than, say, lexical
+variables.
 
-    db DBIx::Squirrel->connect($dsn, $user, $pass, \%attr);
+Helper semantics deal with three common types of interaction:
 
-    # Use the "db" proxy function to associate the "named_foo"
-    # and "all_foo" proxy functions with prepared statements...
+=over
 
-    named_foo db->prepare("select * from foo where name=?");
-    all_foo   db->prepare("select * from foo");
+=item * B<Establishing an association>
 
-    # Use statement and iterator proxy functions (without passing
-    # arguments) just as you would normal scalar object references...
+Before it can be used, a helper must first be associated with a database
+entity. This accomplished by passing the function single argument: a
+reference to the associated object.
 
-    named_foo->execute("baz");
-    all_foo->execute;
+Once established, associations are I<sticky> and cannot easily be undone.
+You should take care to create them once only, in a sensible place.
 
-    # Statement and iterator proxy functions will call their
-    # object's "execute" method when arguments (bind values)
-    # are passed.
+Use Perl's standard importing mechanisms (as shown above) to share
+associations among different package namespaces.
+
+=item * B<Resolving an association>
+
+Fetching the reference to the associated database object is accomplished
+by calling the helper function without any arguments.
+
+When no association exists in this scenario, a helper returns C<undef>.
+
+=item * B<Addressing an association>
+
+Addressing an association amounts to doing something meaningful with it.
+
+We do this by calling the helper function with one or more arguments. Once
+associated with a database object, a helper function will any arguments
+that are passed to it and send a version of these to the database object
+method that imbues meaning to the interaction.
+
+Meaning in this context is determined by the type of association:
+
+=over
+
+=item * 
+
+for a database connection, a statement is prepared using the C<prepare> method;
+
+=item *
+
+for statements and iterators, these are executed with the C<execute> and C<iterate>
+methods respectively.
+
+=back
+
+Clearly there is a paradox here, and it centres around statements expect no
+bind-values.
+
+Optionally, you may enclose any arguments inside anonymous array or hash. In
+order to coerce the helper into performing the execution, you are allowed to
+pass an empty array reference (C<[]>) or hash reference (C<{}>), or resolve
+the association and call the relevant method manually.
+
+=back
+
+=head4 Examples
+
+=over
+
+=item 1.
+
+Let us do a full worked example. We will connect to a database, create and
+work with two result sets, one of which expects a single bind-value. Some
+concepts will be expanded upon later, but it might be helpful to dip a
+toe in the water ahead of time:
+
+    use DBIx::Squirrel database_objects => [ qw/db artists artist/ ];
+
+    # Associate helper ("db") with our database connection:
+
+    @connect_args = ( 'dbi:SQLite:dbname=chinook.db', '', '', { sqlite_unicode => 1 } );
+    db( DBIx::Squirrel->connect(@connection_args) );
+
+    # Resolve the database connection helper ("db"), using it to
+    # associate helpers ("artist" and "artists") with different
+    # result sets:
+
+    artist( db->results('SELECT * FROM artists WHERE Name=? LIMIT 1') );
+    artists( db->results('SELECT * FROM artists') );
+
+    # Address the helper ("artist"), passing it a bind-value, to get
+    # the ArtistId of the artist whose name is "Aerosmith".
     #
-    # A set of bind values may also be passed inside an anonymous
-    # list or hash, and this can be left empty to force execution
-    # of any statement that doesn't take parameters.
+    # We could have called "next" to get the only matching record, but by
+    # calling "single" (or "first") we can ensure that there are no warnings
+    # about dangling active statements emitted when we disconnect from the
+    # database.
 
-    named_foo ["baz"];
-    all_foo   [];
+    print artist('Aerosmith')->single->ArtistId, "\n";
 
-=head3 Database connection
+    # Iterate over the "artists" result set, printing the Name-column for
+    # each artist:
+
+    while ( artists->next ) {
+        print $_->Name, "\n";
+    };
+
+
+=back
+
+=head2 Database connection
 
 Connecting to a database using C<DBIx::Squirrel> works the same
 way as it does when using the C<DBI> C<connect> and C<connect_cached>
@@ -433,7 +513,7 @@ accept a database handle in place of a datasource name. The database
 handle can even be a reference to a C<DBI> object. The original database
 connection will be cloned as as C<DBIx::Squirrel> object.
 
-=head3 Statement preparation
+=head2 Statement preparation
 
 =over
 
@@ -456,7 +536,7 @@ any of these style will work regardless of the driver in use.
 
 =back
 
-=head3 Results processing
+=head2 Results processing
 
 =over
 
